# PWBus - _Connector Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import traceback
from connection_pool import ConnectionPool

from pwbus.commons.logging import *
from pwbus.connectors._connection_pool import _ConnectionPool

_connection_pool = _ConnectionPool()

# _Connector
#
#


class _Connector:

    def __init__(self, connector_type, host, port):
        self.connector_type = connector_type
        self.host = host if host else 'none'
        self.port = port if port else '0'
        self.connector_type_key = f'{self.connector_type}!!{self.host}:{str(self.port)}'
        self.connection_pool = _connection_pool

    def getType(self): return self.connector_type
    def getHost(self): return self.host
    def getPort(self): return self.port

    def getConnectionPool(self):
        if self.connection_pool.exists(self.connector_type_key):
            return self.connection_pool.get(self.connector_type_key)
        else:
            return None

    def setConnectionPool(self, create_connection, max_size, max_usage, idle, ttl):
        try:
            if not self.connection_pool.exists(self.connector_type_key):
                print(
                    f'🟡 _Connector.setConnectionPool - New connection pool [{self.connector_type_key}]')
                self.connection_pool.update(
                    connector_type=self.connector_type_key,
                    connection_pool=ConnectionPool(
                        create=create_connection,
                        max_size=max_size,
                        max_usage=max_usage,
                        idle=idle,
                        ttl=ttl
                    )
                )
        except:
            print('_Connector.setConnectionPool - Failed to set connection pool')
            traceback.print_exc()


# ConnectorIn
#
#
class ConnectorIn(_Connector):

    def __init__(self, connector_type, host, port):
        super().__init__(connector_type=connector_type, host=host, port=port)


# ConnectorOut
#
#
class ConnectorOut(_Connector):

    def __init__(self, connector_type, host, port):
        super().__init__(connector_type=connector_type, host=host, port=port)



# PWBus - Transformation Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sun Nov 17 14:07:58 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.transformations._transform_json import _TransformJSON
from pwbus.transformations._transform_csv import _TransformCSV


class Transformation():

    # Transformation.execute_in
    #
    def execute_in(self, registry, data):

        # TODO Dynamic Class
        if registry["flow.in.payload_format"] == 'csv':
            transform = _TransformCSV()
        elif registry["flow.in.payload_format"] == 'json':
            transform = _TransformJSON()
        else:
            log_debug(
                f'Transformation.execute - no transformation found for [{registry["flow.in.payload_format"]}]')
            return data

        if transform:
            return transform.parse_in(data)

        if registry['engine.debug']:
            log_debug(f'Transformation.execute - no transformation')

        return data

    # Transformation.execute_out
    #
    def execute_out(self, registry, data):

        # TODO Dynamic Class
        if registry["flow.out.payload_format"] == 'csv':
            transform = _TransformCSV()
        elif registry["flow.out.payload_format"] == 'json':
            transform = _TransformJSON()
        else:
            log_debug(
                f'Transformation.execute - no transformation found for [{registry["flow.out.payload_format"]}]')
            return data

        if transform:
            return transform.parse_out(data)

        return data



# PWBus - _Engine Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import string
import secrets
import re
import traceback

from pwbus.commons.logging import *

# _Engine
#
#


class _Engine():

    def __init__(self, name):
        self.name = name
        self.server_list = []

    # _Engine.run
    #
    def run(self, thread, channel_registry):
        start_threads = channel_registry["engine.start_threads"] \
            if channel_registry["engine.start_threads"] else 2
        running_threads = []
        alphabet = string.ascii_lowercase + string.digits
        try:
            for count in range(start_threads):
                hashcode = ''.join(secrets.choice(alphabet) for i in range(6))
                thread_name = f'{self.name}-{hashcode}-{str(count)}'

                log_debug(f'EngineManager.action - thread [{thread_name}]')
                self.server_list.append(
                    thread(
                        thread_id=count,
                        name=thread_name,
                        channel_registry=channel_registry
                    )
                )

            for created_thread in self.server_list:
                created_thread.start()
                running_threads.append(created_thread)

            return running_threads

        except KeyboardInterrupt:
            log_debug('_Engine.run - Please wait, stopping engines threads')

            for started_thread in running_threads:
                started_thread.do_run = False
                log_debug(
                    F'_Engine.run - Stopping thread [{started_thread.name}]')

            log_debug("_Engine.run - Interrupted by user")
            return None

        except:
            log_error(traceback, '_Engine.run - Error running engines threads')

    # _Engine.serve
    #
    def serve(self, channel_registry):
        raise NotImplementedError("_Engine.serve - Method need be implemented")



# PWBus - SqsFlowIn & SqsFlowOut Classes
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Dec  7 09:46:09 -03 2019

from time import sleep
import traceback

from pwbus.commons.logging import *
from pwbus.flows._flow import FlowIn, FlowOut
from pwbus.connectors.sqs import SqsConnectorIn, SqsConnectorOut

# SqsFlowIn
#
#


class SqsFlowIn(FlowIn):

    def __init__(self, request=None, host=None, port=None):
        super().__init__('in', SqsConnectorIn(host=host, port=port))
        self.host = host
        self.port = port

    # SqsFlowIn.receive
    #
    def receive(self, timeout=20):
        DEBUG = self.isDebugEnabled()
        resource_name = self.getRecvFrom()

        if DEBUG:
            log_debug(
                f'SqsFlowIn.receive - Started to [{resource_name}] format={self.getPayloadFormat()}')

        while True:
            try:
                if DEBUG:
                    log_debug("SqsFlowIn.receive.message - Getting connection")

                with self.getConnector().getConnectionPool().item() as connection:

                    # pooling for wait messages
                    while True:

                        if DEBUG:
                            log_debug(
                                f'⏱️  SqsFlowIn.receive.message - Waiting for messages from [{resource_name}|PSEUDOSYNC] ...')

                        # Long poll for message on provided SQS queue
                        received_messages = connection.receive_message(
                            QueueUrl=resource_name,
                            AttributeNames=[
                                'SentTimestamp'
                            ],
                            MaxNumberOfMessages=1,
                            MessageAttributeNames=[
                                'All'
                            ],
                            WaitTimeSeconds=timeout,
                            VisibilityTimeout=5
                        )

                        if 'Messages' in received_messages:
                            message = received_messages['Messages'][0]
                            payload_in = message['Body']

                            connection.delete_message(
                                QueueUrl=resource_name,
                                ReceiptHandle=message['ReceiptHandle']
                            )

                            if DEBUG:
                                log_debug(
                                    f'↘️  SqsFlowIn.receive.message - Message received successfully from [{resource_name}] - payload_in [{payload_in}]')
                            return payload_in

            except KeyboardInterrupt:
                log_debug('SqsFlowIn.receive.message - Interrupted by user')
                self.stop()

            except:
                log_error(
                    traceback, f'SqsFlowIn.receive.message - Failed on receive message from [{self.getRecvFrom()}]')
                sleep(5)

# SqsFlowOut
#
#


class SqsFlowOut(FlowOut):

    def __init__(self, host=None, port=None):
        super().__init__('out', SqsConnectorOut(host=host, port=port))
        self.host = host
        self.port = port

    # SqsFlowOut.send
    #
    def send(self):
        DEBUG = self.isDebugEnabled()

        resource_name = self.getSendTo()
        payload_out = self.getMessage().getPayloadAsString()

        if DEBUG:
            log_debug(
                f'SqsFlowOut.send - Started to [{resource_name}] - payload_out [{payload_out}]')

        try:
            with self.getConnector().getConnectionPool().item() as connection:

                # Send message to SQS queue
                response = connection.send_message(
                    QueueUrl=resource_name,
                    DelaySeconds=0,
                    MessageAttributes={
                        'Author': {
                            'DataType': 'String',
                            'StringValue': 'Pwbus-Engine-v1'
                        },
                        'Pwbus-Correlation-Id': {
                            'DataType': 'String',
                            'StringValue': self.getMessage().getCorrelationId()
                        }
                    },
                    MessageBody=(
                        payload_out
                    )
                )

                # when fifo
                # MessageGroupId='1'
                # MessageDeduplicationId=self.getMessage().getMessageId()

                if DEBUG:
                    log_debug(
                        f'↗️  SqsFlowOut.send - Sqs MessageId [{response["MessageId"]}]')
                return True
        except:
            log_error(
                traceback, f'SqsFlowOut.send - Failed to send the message to [{resource_name}]')
            traceback.print_exc()

        return False



# PWBus - KafkaEngineThread Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Wed Dec 11 08:22:50 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.engines._engine_thread import _EngineThread


class KafkaEngineThread(_EngineThread):

    def __init__(self, thread_id, name, channel_registry):
        super().__init__(thread_id, name, channel_registry)



### PWBus - _ConnectionPool Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Fri Dec 13 19:17:38 -03 2019

### _ConnectionPoolState
#
class _ConnectionPoolState:
    _shared_state = {}
    def __init__(self):
        self.__dict__ = self._shared_state

### _ConnectionPool (Singleton)
#
class _ConnectionPool( _ConnectionPoolState ):
    def __init__(self):
        _ConnectionPoolState.__init__(self)
        self.state = dict()

    def __str__(self): return self.state

    def get(self, key):
        return self.state[key] \
               if key in self.state else None

    def update(self, connector_type, connection_pool):
        self.state.update({connector_type: connection_pool})

    def exists(self, connector_type):
        return True if connector_type in self.state else False



# PWBus - KafkaEngine Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Wed Dec 11 08:22:50 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.engines._engine import _Engine
from pwbus.engines.kafka.engine_thread import KafkaEngineThread


class KafkaEngine(_Engine):

    def __init__(self):
        super().__init__(
            name='kafka'
        )

    # KafkaEngine.serve
    #
    def serve(self, channel_registry):
        return self.run(KafkaEngineThread, channel_registry)



# PWBus - _TransformCSV Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sun Nov 17 14:07:58 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.transformations._transform import _Transform


class _TransformCSV(_Transform):

    # _TransformCSV.parsein - return Dict
    #
    def parse_in(self, payload):
        keys = []
        values = dict()

        for line in payload.splitlines():
            if not keys:

                for key in line.split(';'):
                    keys.append(key)
            else:
                i = 0
                for value in line.split(';'):
                    key = keys[i]
                    if key == '':
                        continue
                    values.update({key: value})
                    i += 1

        return values

    # _TransformCSV.parse_out - receive Dict
    #
    def parse_out(self, payload):
        payload_out = ''

        for key in payload:
            payload_out += f'{key};'
        payload_out += "\n"

        for key in payload:
            payload_out += f'{payload[key]};'
        payload_out += "\n"

        return payload_out



# PWBus - EngineDispatch Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sun Nov 17 07:25:07 -03 2019

from time import sleep
import importlib
import traceback

from pwbus.commons.logging import *
from pwbus.transformations.transformation import Transformation
from pwbus.tasks.dynamic_task import DynamicTask
from pwbus.engines.engine_monitor_event import EngineMonitorEvent


# EngineDispatch
#
#
class EngineDispatch():

    # RedisEngine.serve
    #
    def route(self, channel_registry, request=None):

        try:
            if not channel_registry or not channel_registry['channel']:
                log_debug('EngineDispatch.route - registry is invalid')
                return None

            DEBUG = channel_registry['engine.debug']

            if DEBUG:
                log_debug(
                    f'EngineDispatch.route - Message dispatch started for [{channel_registry["channel"]}]')

            # @@@ step 1 - retrieve message in

            # Flow input

            # Get FlowIn Class dynamically
            resource_type = channel_registry["flow.in.resource_type"].lower()
            module_name = f'pwbus.flows.{resource_type}'
            class_name = f'{resource_type.capitalize()}FlowIn'
            module = importlib.import_module(module_name)
            class_ = getattr(module, class_name)
            flow_in = class_(
                request=request,
                host=channel_registry["flow.in.host"] if "flow.in.host" in channel_registry else None,
                port=channel_registry["flow.in.port"] if "flow.in.port" in channel_registry else None
            )

            flow_in.setRegistry(channel_registry)

            # receive payload
            payload_in = flow_in.receive()
            if payload_in is None:
                flow_in.getMessage().setHeadersEntry("Pwbus-Status-Code", 500)
                flow_in.getMessage().setHeadersEntry("Pwbus-Info-Message",
                                                     "Failed on receive request message")
                EngineMonitorEvent().incrValue("pwbus_flow_errors")
                EngineMonitorEvent().incrValue("pwbus_transformation_errors")
                sleep(2)
                return None

            # start racing
            start_racing = getMillis()
            if DEBUG:
                log_debug(f'▶️  EngineDispatch.route - Start message flow')

            # transform payload in
            transformation = Transformation()
            transformed_payload = transformation.execute_in(
                channel_registry,
                payload_in
            )

            # prepare message in
            flow_in.prepareMessage(transformed_payload)

            # @@@ step 2 - execute task
            if flow_in.isTaskEnabled():
                field_task_id = flow_in.getMessage().getTaskId()
                if field_task_id is not None:
                    if DEBUG:
                        log_debug(
                            f'EngineDispatch - Flow running for task_id [{field_task_id}]')

                    try:
                        for task_id in field_task_id.split(','):
                            task = DynamicTask(
                                task_id, transformed_payload, DEBUG)
                            task.getInstance().execute()
                    except:
                        EngineMonitorEvent().incrValue("pwbus_task_errors")
                        pass

            # @@@ step 3 - send message out

            # Get FlowOut Class dynamically
            resource_type = channel_registry["flow.out.resource_type"].lower()
            module_name = f'pwbus.flows.{resource_type}'
            class_name = f'{resource_type.capitalize()}FlowOut'
            module = importlib.import_module(module_name)
            class_ = getattr(module, class_name)
            flow_out = class_(
                host=channel_registry["flow.out.host"] if "flow.out.host" in channel_registry else None,
                port=channel_registry["flow.out.port"] if "flow.out.port" in channel_registry else None
            )

            # prepare message out
            flow_out.setResponse(flow_in.getResponse())
            flow_out.setHeaders(flow_in.getHeaders())
            message = flow_in.getMessage()

            if transformed_payload is None:
                if DEBUG:
                    log_debug(
                        f'🟥 EngineDispatch.route - Failed on transform response message')
                EngineMonitorEvent().incrValue("pwbus_flow_errors")
                return None

            if DEBUG:
                log_debug(
                    f'EngineDispatch.route - transformed_payload=[{transformed_payload}]')

            # Move headers entries to payload
            headers = flow_in.getHeaders()
            for field in headers:
                transformed_payload[field] = headers[field]

            message.setPayload(transformed_payload)

            # transform payload out
            transformation = Transformation()
            transformed_payload = transformation.execute_out(
                channel_registry,
                message.getPayload()
            )

            message.setPayloadAsString(transformed_payload)

            # send message out
            flow_out.setRegistry(channel_registry)
            flow_out.prepareMessage(message, headers)

            if channel_registry["flow.message_dump"]:
                log_message_dump('EngineDispatch.route', transformed_payload)

            # end message transformation
            end_transform = getMillis()

            if flow_out:
                if not flow_out.send():
                    if DEBUG:
                        log_debug(
                            f'🟥 EngineDispatch.route ({channel_registry["flow.out.resource_type"]}) - Failed on send response message - Waiting 10s...')
                    EngineMonitorEvent().incrValue("pwbus_flow_errors")
                    sleep(2)
                    return None

            # end racing
            end_racing = getMillis()

            if DEBUG:
                log_debug(
                    f'✅ EngineDispatch.route - End message flow - Completed in {end_racing-start_racing}ms - delivery {(end_racing-start_racing)-(end_transform-start_racing)}ms - correlation_id [{flow_in.getMessage().getCorrelationId()}]\n\n')

            EngineMonitorEvent().incrValue("pwbus_flow_success")
            return flow_out

        except InterruptedError:
            log_debug(f'EngineDispatch.route - Stopping flow')
            return None

        except:
            traceback.print_exc()
            log_error(
                traceback, f'EngineDispatch.route - Error in flow execution')



# PWBus - RabbitmqFlowIn & RabbitmqFlowOut Classes
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import pika
import traceback

from pwbus.commons.logging import *
from pwbus.flows._flow import FlowIn, FlowOut
from pwbus.connectors.rabbitmq import RabbitmqConnectorIn, RabbitmqConnectorOut

# RabbitmqFlowIn
#
#


class RabbitmqFlowIn(FlowIn):

    def __init__(self, request=None, host=None, port=None):
        super().__init__('in', RabbitmqConnectorIn(host=host, port=port))
        self.host = host
        self.port = port
        self.payload_in = None
        self.mqchannel = None

    # RabbitmqFlowIn.receive
    #
    def receive(self, timeout=None):
        DEBUG = self.isDebugEnabled()
        resource_name = self.getRecvFrom()

        if DEBUG:
            log_debug(
                f'RabbitmqFlowIn.receive - Started to [{resource_name}] format={self.getPayloadFormat()}')

        try:
            if DEBUG:
                log_debug("RabbitmqFlowIn.receive.message - Getting connection")

            with self.getConnector().getConnectionPool().item() as connection:

                if not connection:
                    return None
                mqchannel = connection.channel()
                self.setMqchannel(mqchannel)

                mqchannel.queue_declare(queue=resource_name, durable=True, arguments={
                                        "x-queue-type": "classic"})
                mqchannel.basic_consume(
                    queue=resource_name,
                    auto_ack=True,
                    on_message_callback=self.receive_callback
                )

                if DEBUG:
                    log_debug(
                        f'⏱️  RabbitmqFlowIn.receive.message - Waiting for messages from [{resource_name}] ...')
                mqchannel.start_consuming()

                # convert rabbitmq payload from tuple type
                payload_in = self.getPayloadIn()

                if payload_in:
                    if DEBUG:
                        log_debug(
                            f'↘️  RabbitmqFlowIn.receive.message - Message received successfully from [{resource_name}] - payload_in [{payload_in}]')
                    return ''.join(payload_in.decode('utf-8'))

        except KeyboardInterrupt:
            log_debug('RabbitmqFlowIn.receive.message - Interrupted by user')
            self.stop()

        except:
            log_error(
                traceback, f'RabbitmqFlowIn.receive.message - Failed on receive message from [{self.getRecvFrom()}]')

        return None

    # RabbitmqFlowIn.setConnection
    #
    def setMqchannel(self, mqchannel):
        self.mqchannel = mqchannel

    # RabbitmqFlowIn.getConnection
    #
    def getMqchannel(self):
        return self.mqchannel

    # RabbitmqFlowIn.getPayloadIn
    #
    def getPayloadIn(self):
        return self.payload_in

    # RabbitmqFlowIn.setPayloadIn
    #
    def setPayloadIn(self, payload_in):
        self.payload_in = payload_in

    def receive_callback(self, ch, method, props, body):
        self.getMqchannel().stop_consuming()
        self.getMessage().setCorrelationId(props.correlation_id)
        self.setPayloadIn(body)


# RabbitmqFlowOut
#
#
class RabbitmqFlowOut(FlowOut):

    def __init__(self, host=None, port=None):
        super().__init__('out', RabbitmqConnectorOut(host=host, port=port))
        self.host = host
        self.port = port

    # RabbitmqFlowOut.send
    #
    def send(self):
        DEBUG = self.isDebugEnabled()

        resource_name = self.getSendTo()
        payload_out = self.getMessage().getPayloadAsString()

        if DEBUG:
            log_debug(
                f'RabbitmqFlowOut.send - Started to [{resource_name}] - payload_out [{payload_out}]')

        try:
            with self.getConnector().getConnectionPool().item() as connection:
                if not connection:
                    return False
                mqchannel = connection.channel()
                mqchannel.queue_declare(queue=resource_name, durable=True, arguments={
                                        "x-queue-type": "classic"})
                #reply_to=resource_name.replace('request', 'response'),
                mqchannel.basic_publish(
                    exchange='',
                    routing_key=resource_name,
                    properties=pika.BasicProperties(
                        correlation_id=self.getMessage().getCorrelationId()
                    ),
                    body=payload_out
                )

                log_debug(
                    f'↗️  RabbitmqFlowOut.send - Rabbitmq basic_publish to [{resource_name}] successfully')
                return True

        except:
            log_error(
                traceback, f'RabbitmqFlowOut.send - Failed to send the message to [{resource_name}]')
            traceback.print_exc()

        return False
